#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for a cloud service."""


import os
import threading

import libpal.palcmd as palcmd
from iobase.xmldata import (XmlConfNode, ANY_CHILD_NAME, escape)
from pal_info import (ReqTypes, Results)

from landiskcloud.proxy import (PROXY_NAME, HTTPS_PROTOCOL, ProxyInfo,)
try:
    from landiskaccess.fletsazukeruaccess import (FletsAzukeruAccess,)
    from landiskcloud.info.fletsazukeru import (FletsAzukeruConnectionInfo,
                                            FletsAzukeruTargetInfo)
    from pkgaddon.flets_azukeru_config import (FLETS_AZUKERU_CONF_PATH,)
except:
    pass

try:
    from landiskaccess.amazons3access import (AmazonS3Access,)
    from landiskcloud.info.amazons3 import (AmazonS3ConnectionInfo,)
except:
    pass

try:
    from landiskaccess.nifcloudaccess import (NifCloudAccess,)
    from landiskcloud.info.nifcloud import (NifCloudConnectionInfo,)
except:
    pass

try:
    from ionas.fileaccess.azureaccess import (AzureAccess,)
    from landiskcloud.info.azure import (AzureConnectionInfo,)
except:
    pass

try:
    from landiskaccess.googledriveaccess import (GoogleDriveAccess,)
    from landiskcloud.info.googledrive import (GoogleDriveConnectionInfo,)
    import pkgaddon.googledrive_config as googledrive_config
except:
    pass

try:
    from landiskaccess.dropboxaccess import (DropboxAccess,)
    from landiskcloud.info.dropbox import (DropboxConnectionInfo,)
    import pkgaddon.dropbox_config as dropbox_config
except:
    pass

try:
    from landiskaccess.boxaccess import (BoxAccess,)
    from landiskcloud.info.box import (BoxConnectionInfo,)
    import pkgaddon.box_pkg_config as box_pkg_config
except:
    pass

try:
    from landiskaccess.mdcloudaccess import (MdCloudAccess,)
    from landiskcloud.info.mdcloud import (MdCloudConnectionInfo,)
except:
    pass


URL_XML_TMPL = "<url>%s</url>"

SUCCESS_RESULT = "Success"
FAILED_RESULT = "Failed"
TIMEOUT_RESULT = "TimeOut"
LOGIN_FAILED_RESULT = "LoginFailed"

CHECK_TIMEOUT_SEC = 20
RESULT_KEY = "result"
ROOT_PATH = "/"

OLDNAME_NAME = "oldname"
RESULT_NAME = "result"

FLETSAZUKERU_NAME = "fletsazukeru"
AMAZONS3_NAME = "amazons3"
NIFCLOUD_NAME = "nifcloud"
AZURE_NAME = "azure"
GOOGLEDRIVE_NAME = "googledrive"
DROPBOX_NAME = "dropbox"
BOX_NAME = "box"
MDCLOUD_NAME = "mdcloud"
HTTP_METHOD = "http"


def _exec_check(client, result_dict):
    """Thread main for checking to access to the BOX server."""

    try:
        client._conn.get_all_buckets()
        result_dict[RESULT_KEY] = SUCCESS_RESULT
        return

#    except LoginError:
#        result_dict[RESULT_KEY] = LOGIN_FAILED_RESULT
#
    except:
        result_dict[RESULT_KEY] = FAILED_RESULT

    try:
        client.client.list_containers()
        result_dict[RESULT_KEY] = SUCCESS_RESULT
        return

    except:
        result_dict[RESULT_KEY] = FAILED_RESULT

    try:
        client.stat_directory(ROOT_PATH)
        result_dict[RESULT_KEY] = SUCCESS_RESULT
        return

    except:
        result_dict[RESULT_KEY] = FAILED_RESULT

    try:
        client.get_contents(ROOT_PATH)
        result_dict[RESULT_KEY] = SUCCESS_RESULT
        return

    except:
        result_dict[RESULT_KEY] = FAILED_RESULT


class PalCloud(palcmd.PalCmd):
    """PAL Cloud command implementation class."""

    def __get_handler(self, entry):
        """Handle get request."""

        target_node = entry.get().get_node(ANY_CHILD_NAME)
        if target_node.get_node_name() == GOOGLEDRIVE_NAME:
            authenticate_url = GoogleDriveAccess(
                            googledrive_config.CLIENT_ID,
                            googledrive_config.CLIENT_SECRET,
                            scopes=googledrive_config.SCOPES,
                            redirect_url=googledrive_config.REDIRECT_URL,
                            session_path="",
                            proxy=None).auth_url
        elif target_node.get_node_name() == DROPBOX_NAME:
            authenticate_url = DropboxAccess(
                            dropbox_config.APP_KEY,
                            dropbox_config.APP_SECRET,
                            session_path="",
                            proxy=None).auth_url
        elif target_node.get_node_name() == BOX_NAME:
            user_agent = box_pkg_config.HTTP_USER_AGENT_TMPL
            authenticate_url = BoxAccess(
                            box_pkg_config.CLIENT_ID,
                            box_pkg_config.CLIENT_SECRET,
                            box_pkg_config.REST_BASE_URL,
                            box_pkg_config.UPLOAD_BASE_URL,
                            box_pkg_config.OAUTH_BASE_URL,
                            box_pkg_config.REDIRECT_URL,
                            session_path="",
                            proxy=None,
                            http_user_agent=user_agent).auth_url
        else:
            authenticate_url = "http://dummy/dummy"
        target_node.append_child(XmlConfNode(URL_XML_TMPL %
                                            escape(authenticate_url)))
        entry.set_result(Results.OK)
        return entry

    def __get_session_path(self, target_node, connection_name=""):
        """Return session_path if target use it."""

        session_dir = ""
        connection_info = None
        if target_node.get_node_name() == GOOGLEDRIVE_NAME:
            connection_info = GoogleDriveConnectionInfo(target_node)
            session_dir = googledrive_config.SESSION_DIR

        elif target_node.get_node_name() == DROPBOX_NAME:
            connection_info = DropboxConnectionInfo(target_node)
            session_dir = dropbox_config.SESSION_DIR

        elif target_node.get_node_name() == BOX_NAME:
            connection_info = BoxConnectionInfo(target_node)
            session_dir = box_pkg_config.SESSION_DIR

        else:
            return ""

        if not connection_name and connection_info is not None:
            connection_name = connection_info.name

        if session_dir and connection_name:
            return os.path.join(session_dir, connection_name) + ".session"

        return ""

    def __get_client(self, entry):
        """Return client class instance for target_node."""

        target_node = entry.get().get_node(ANY_CHILD_NAME)
        if target_node.get_node_name() == FLETSAZUKERU_NAME:
            connection_info = FletsAzukeruConnectionInfo(target_node)

            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy = proxy_info.get_proxy_url(HTTPS_PROTOCOL)
            except:
                proxy = None

            return FletsAzukeruAccess(
                    connection_info.clientid,
                    connection_info.password,
                    connection_info.host_name,
                    connection_info.company_id,
                    connection_info.service_id,
                    FletsAzukeruTargetInfo.MY_FOLDER,
                    terminal_id="",
                    terminal_key=connection_info.terminalkey,
                    using_image_login=connection_info.using_image_login,
                    conf_path=FLETS_AZUKERU_CONF_PATH,
                    proxy=proxy)
        elif target_node.get_node_name() == AMAZONS3_NAME:
            connection_info = AmazonS3ConnectionInfo(target_node)

            proxy_host = None
            proxy_port = None
            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy_host = proxy_info.host
                proxy_port = proxy_info.port
            except:
                pass

            is_secure = True
            if connection_info.httpmethod == HTTP_METHOD:
                is_secure = False

            return AmazonS3Access(
                    connection_info.accesskey,
                    connection_info.secretkey,
                    None,
                    connection_info.endpoint,
                    is_secure=is_secure,
                    proxy=proxy_host,
                    proxy_port=proxy_port)
        elif target_node.get_node_name() == NIFCLOUD_NAME:
            connection_info = NifCloudConnectionInfo(target_node)

            proxy_host = None
            proxy_port = None
            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy_host = proxy_info.host
                proxy_port = proxy_info.port
            except:
                pass

            is_secure = True
            if connection_info.httpmethod == HTTP_METHOD:
                is_secure = False

            return NifCloudAccess(
                    connection_info.accesskey,
                    connection_info.secretkey,
                    None,
                    connection_info.endpoint,
                    is_secure=is_secure,
                    proxy=proxy_host,
                    proxy_port=proxy_port)
        elif target_node.get_node_name() == AZURE_NAME:
            connection_info = AzureConnectionInfo(target_node)

            proxy_host = None
            proxy_port = None
            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy_host = proxy_info.host
                proxy_port = proxy_info.port
            except:
                pass

            return AzureAccess(
                    connection_info.accountname,
                    connection_info.accesskey,
                    None,
                    connection_info.httpmethod,
                    proxy_host=proxy_host,
                    proxy_port=proxy_port)

        elif target_node.get_node_name() == GOOGLEDRIVE_NAME:
            connection_info = GoogleDriveConnectionInfo(target_node)

            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy = proxy_info.get_proxy_url(HTTPS_PROTOCOL)
            except:
                proxy = None

            session_path = self.__get_session_path(target_node)
            client = GoogleDriveAccess(
                        googledrive_config.CLIENT_ID,
                        googledrive_config.CLIENT_SECRET,
                        scopes=googledrive_config.SCOPES,
                        redirect_url=googledrive_config.REDIRECT_URL,
                        session_path=session_path,
                        proxy=proxy)
            if not os.path.exists(session_path):
                client.authenticate({"code": connection_info.code})
            else:
                client.authenticate()
            return client

        elif target_node.get_node_name() == DROPBOX_NAME:
            connection_info = DropboxConnectionInfo(target_node)

            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy = proxy_info.host
                if proxy_info.port:
                    proxy += ":" + str(proxy_info.port)
            except:
                proxy = None

            session_path = self.__get_session_path(target_node)
            client = DropboxAccess(
                        dropbox_config.APP_KEY,
                        dropbox_config.APP_SECRET,
                        session_path=session_path,
                        proxy=proxy)
            if not os.path.exists(session_path):
                client.authenticate({"code": connection_info.code})
            else:
                client.authenticate()
            return client
        elif target_node.get_node_name() == BOX_NAME:
            connection_info = BoxConnectionInfo(target_node)

            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy = proxy_info.get_proxy_url(HTTPS_PROTOCOL)
            except:
                proxy = None

            session_path = self.__get_session_path(target_node)
            client = BoxAccess(
                        box_pkg_config.CLIENT_ID,
                        box_pkg_config.CLIENT_SECRET,
                        box_pkg_config.REST_BASE_URL,
                        box_pkg_config.UPLOAD_BASE_URL,
                        box_pkg_config.OAUTH_BASE_URL,
                        box_pkg_config.REDIRECT_URL,
                        session_path=session_path,
                        proxy=proxy,
                        http_user_agent=box_pkg_config.HTTP_USER_AGENT_TMPL)

            if not os.path.exists(session_path):
                client.authenticate({"code": connection_info.code})
            else:
                client.authenticate()
            return client

        elif target_node.get_node_name() == MDCLOUD_NAME:
            connection_info = MdCloudConnectionInfo(target_node)

            proxy_host = None
            proxy_port = None
            try:
                proxy_info = ProxyInfo(entry.get().get_node(PROXY_NAME))
                proxy_host = proxy_info.host
                proxy_port = proxy_info.port
            except:
                pass

            is_secure = True
            if connection_info.httpmethod == HTTP_METHOD:
                is_secure = False

            return MdCloudAccess(
                    connection_info.accesskey,
                    connection_info.secretkey,
                    None,
                    connection_info.endpoint,
                    is_secure=is_secure,
                    proxy=proxy_host,
                    proxy_port=proxy_port)

    def __add_handler(self, entry):
        """Handle add request."""

        try:
            target_node = entry.get().get_node(ANY_CHILD_NAME)
            session_path = self.__get_session_path(target_node)
            if os.path.exists(session_path):
                self.error("Add Failed - session file already exists!!")
                entry.set_result(Results.ADD_ERROR)
                return entry

            self.__get_client(entry)
            entry.set_result(Results.OK)
        except:
            self.exception("Add Failed!!")
            entry.set_result(Results.ADD_ERROR)

        return entry

    def __modify_handler(self, entry):
        """Handle modify request."""

        try:
            target_node = entry.get().get_node(ANY_CHILD_NAME)
            session_path = self.__get_session_path(target_node)
            old_name = target_node.get_value(OLDNAME_NAME, "")
            if session_path and old_name:
                old_session_path = \
                            self.__get_session_path(target_node, old_name)
                os.rename(old_session_path, session_path)
            entry.set_result(Results.OK)
        except:
            self.exception("Modify Failed!!")
            entry.set_result(Results.MODIFY_ERROR)

        return entry

    def __delete_handler(self, entry):
        """Handle delete request."""

        try:
            target_node = entry.get().get_node(ANY_CHILD_NAME)
            session_path = self.__get_session_path(target_node)
            if session_path and os.path.exists(session_path):
                os.unlink(session_path)
            entry.set_result(Results.OK)
        except:
            self.exception("Delete Failed!!")
            entry.set_result(Results.DELETE_ERROR)

        return entry

    def __check_handler(self, entry):
        """Handle check request."""

        result = SUCCESS_RESULT
        try:
            client = self.__get_client(entry)
            result_dict = {}
            check_thread = threading.Thread(target=_exec_check,
                                            args=(client, result_dict))
            check_thread.daemon = True
            check_thread.start()
            check_thread.join(CHECK_TIMEOUT_SEC)

            if check_thread.is_alive():
                result = TIMEOUT_RESULT
            else:
                result = result_dict[RESULT_KEY]
        except:
            self.exception("Check Failed!!")
            result = FAILED_RESULT

        if result == SUCCESS_RESULT:
            entry.set_result(Results.OK)
        else:
            entry.set_result(Results.CHECK_ERROR)

        entry.get().append_child(RESULT_NAME, result)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.GET, self.__get_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)

#
# Main
#

if __name__ == "__main__":
    PalCloud(option=palcmd.PalCmdOption()).handle_request()
