#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Fuji Xerox Working Folder Settings."""

import time
import urllib2

from iobase.exception import InvalidArgError
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.xmldata import XmlConfNode

import libpal.palcmd as palcmd
import libpal.util as util

from pal_info import (ReqTypes, Results)

from fxwf.fxwfinfo import FxwfInfo
from fxwf.fxwfclient import FxwfClient
from pkgaddon.fxwf_config import (REST_BASE_URL, CLIENT_ID, CLIENT_SECRET)
from pkgaddon.pal_fxwf_config import FXWFSYNC_CMD


X_SHARE = "share"
X_NAME = "name"
X_SHARESERVICE = "shareservice"
X_SYNC = "fxwfsync"
X_STATUS = "status"
X_CABINET = "cabinetnumber"
X_ACCOUNT = "userid"
X_PASSWD = "passwd"
X_DRAWERLIST = "drawerlist"
X_DRAWER = "drawer"
X_DRAWERNAME = "name"
X_DRAWERDESC = "desc"

X_PROXY = "proxy"

X_SHARE_NAME = X_SHARE + "/" + X_NAME
X_SHARE_SYNC = X_SHARE + "/" + X_SHARESERVICE + "/" + \
               X_SYNC + "[@enable='true']"

K_STAT = "stat"
K_NAME = "name"
K_DESC = "description"

SYNC_CMD_OPTION_INFO = "--info"
SYNC_CMD_OPTION_FORCE_DOWNLOAD = "--force-download"

SYNC_CMD_OUTPUT_STATUS_NOERROR = "status: noerror"

MAX_RETRY = 1


class PalFxwf(palcmd.PalCmd):
    """Pal Fxwf command implementation class."""

    #
    # Definitions
    #

    STATUS_NOERROR = "noerror"
    STATUS_FAIL = "fail"

    #
    # Internal methods
    #

    def __execute_cloud_sync_command(
            self, command, option, share_name, stdout):
        """Execute cloud sync command with specified option."""

        if (option != SYNC_CMD_OPTION_INFO and
            option != SYNC_CMD_OPTION_FORCE_DOWNLOAD):
            raise InvalidArgError("Invalid option argument.")

        cmdline = [command, option, share_name]
        return CmdExec(cmdline, stdout=stdout, stderr=NULL).execute()

    def __list_drawers(self, session):
        """List drawers with retry."""

        for count in range(0, MAX_RETRY):
            try:
                return session.list_drawers()
            except urllib2.HTTPError:
                if count == MAX_RETRY - 1:
                    raise
                time.sleep(1)
            except:
                raise

    def __close_session(self, session):
        """Close session with retry."""

        for count in range(0, MAX_RETRY):
            try:
                session.close_session()
                break
            except urllib2.HTTPError:
                if count == MAX_RETRY - 1:
                    raise
                time.sleep(1)
            except:
                raise

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle query request."""

        try:
            src_root_node = entry.get()
            src_share_name = src_root_node.get_value(X_SHARE_NAME)
            src_root_node.get_node(X_SHARE_SYNC)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        try:
            dst_entry = entry.get_clone()
            dst_root_node = dst_entry.get()
            dst_sync_node = dst_root_node.get_node(X_SHARE_SYNC)
            dst_sync_node.append_child(X_STATUS, text=self.STATUS_FAIL)
        except:
            dst_entry.set_result(Results.QUERY_ERROR)
            return dst_entry

        try:
            proc = self.__execute_cloud_sync_command(
                FXWFSYNC_CMD, SYNC_CMD_OPTION_INFO,
                src_share_name, PIPE)
            for line in proc.stdout.readlines():
                if line.startswith(SYNC_CMD_OUTPUT_STATUS_NOERROR):
                    dst_status_node = dst_sync_node.get_node(X_STATUS)
                    dst_status_node.set_text(self.STATUS_NOERROR)
        except:
            dst_entry.set_result(Results.QUERY_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __list_handler(self, entry):
        """Handle list request."""

        try:
            src_root_node = entry.get()
            src_sync_node = src_root_node.get_node(X_SYNC)
            src_cabinet = src_sync_node.get_value(X_CABINET)
            src_account = src_sync_node.get_value(X_ACCOUNT)
            src_passwd = src_sync_node.get_value(X_PASSWD)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        proxies = None
        try:
            src_proxy_node = src_root_node.get_node(X_PROXY)
            proxies = util.get_proxies(src_proxy_node)
        except:
            pass

        try:
            dst_entry = entry.get_clone()
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry

        session = None
        dst_drawer_list = []
        try:
            session = FxwfClient(
                FxwfInfo(
                    REST_BASE_URL, src_cabinet, src_account, src_passwd,
                    CLIENT_ID, CLIENT_SECRET),
                proxies=proxies)
            session.open_session()

            for drawer in self.__list_drawers(session):
                dst_drawer = {K_NAME: drawer[K_NAME], K_DESC: drawer[K_DESC]}
                dst_drawer_list.append(dst_drawer)
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry
        finally:
            try:
                self.__close_session(session)
            except:
                dst_entry.set_result(Results.LIST_ERROR)
                return dst_entry

        try:
            dst_root_node = dst_entry.get()
            dst_sync_node = dst_root_node.get_node(X_SYNC)

            dst_drawerlist_node = XmlConfNode(X_DRAWERLIST)
            for dst_drawer in dst_drawer_list:
                try:
                    dst_drawer_node = XmlConfNode(X_DRAWER)
                    dst_drawer_node.append_child(
                        X_DRAWERNAME, text=dst_drawer[K_NAME])
                    dst_drawer_node.append_child(
                        X_DRAWERDESC, text=dst_drawer[K_DESC])
                    dst_drawerlist_node.append_child(dst_drawer_node)
                except:
                    pass
            dst_sync_node.append_child(dst_drawerlist_node)
        except:
            dst_entry.set_result(Results.LIST_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    def __download_handler(self, entry):
        """Handle download request."""

        try:
            src_root_node = entry.get()
            src_share_name = src_root_node.get_value(X_SHARE_NAME)
            src_root_node.get_node(X_SHARE_SYNC)
        except:
            entry.set_result(Results.INVALID_XML)
            return entry

        try:
            dst_entry = entry.get_clone()
        except:
            dst_entry.set_result(Results.DOWNLOAD_ERROR)
            return dst_entry

        try:
            self.__execute_cloud_sync_command(
                FXWFSYNC_CMD, SYNC_CMD_OPTION_FORCE_DOWNLOAD,
                src_share_name, NULL)
        except:
            dst_entry.set_result(Results.DOWNLOAD_ERROR)
            return dst_entry

        dst_entry.set_result(Results.OK)
        return dst_entry

    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.LIST, self.__list_handler)
        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)


#
# Main
#

if __name__ == "__main__":
    PalFxwf(option=palcmd.PalCmdOption()).handle_request()
